/**
* @file sample_process.cpp
*
* Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/
#include "sample_process.h"
#include <iostream>
#include <vector>
#include <cmath>
#include "preprocess.h"
#include "model_process.h"
#include "postprocess.h"
#include "acl/acl.h"
#include "utils.h"
#include "Utils_Process.h"
#include "lidar_processing.h"
#include <pcl/io/pcd_io.h>
#include <iostream>

using namespace std;
extern bool g_isDevice;


double calculateDistance(const LidarPerceptionOutput& output1, const LidarPerceptionOutput& output2) {
    // Calculate the Euclidean distance between the centers of the two bounding boxes
    // double center1x = (box1.minx + box1.maxx) / 2.0;
    // double center1y = (box1.miny + box1.maxy) / 2.0;
    // double center1z = (box1.minz + box1.maxz) / 2.0;

    // double center2x = (box2.minx + box2.maxx) / 2.0;
    // double center2y = (box2.miny + box2.maxy) / 2.0;
    // double center2z = (box2.minz + box2.maxz) / 2.0;

    double center1x = output1.position.x;
    double center1y = output1.position.y;
    double center1z = output1.position.z;

    double center2x = output2.position.x;
    double center2y = output2.position.y;
    double center2z = output2.position.z;


    return sqrt(pow(center1x - center2x, 2) + pow(center1y - center2y, 2) + pow(center1z - center2z, 2));
}


double calculateIoU(const LidarPerceptionOutput& output1, const LidarPerceptionOutput& output2) {
    // Calculate the intersection and union volumes for IoU
    double intersectionVolume = std::max(0.0, std::min(output1.bbox3d.maxx, output2.bbox3d.maxx) - std::max(output1.bbox3d.minx, output2.bbox3d.minx))
                             * std::max(0.0, std::min(output1.bbox3d.maxy, output2.bbox3d.maxy) - std::max(output1.bbox3d.miny, output2.bbox3d.miny))
                             * std::max(0.0, std::min(output1.bbox3d.maxz, output2.bbox3d.maxz) - std::max(output1.bbox3d.minz, output2.bbox3d.minz));
    
    double unionVolume = (output1.bbox3d.maxx - output1.bbox3d.minx) * (output1.bbox3d.maxy - output1.bbox3d.miny) * (output1.bbox3d.maxz - output1.bbox3d.minz)
                       + (output2.bbox3d.maxx - output2.bbox3d.minx) * (output2.bbox3d.maxy - output2.bbox3d.miny) * (output2.bbox3d.maxz - output2.bbox3d.minz)
                       - intersectionVolume;
    
    // Calculate the IoU
    return (intersectionVolume > 0) ? (intersectionVolume / unionVolume) : 0.0;
}







SampleProcess::SampleProcess() :deviceId_(0), context_(nullptr), stream_(nullptr)
{
}

SampleProcess::~SampleProcess()
{
    DestroyResource();
}

Result SampleProcess::InitResource()
{
    // ACL init
    // acl.json is dump or profiling config file
    if (!Utils::validStatus("aclInit", aclInit(nullptr))) {
        return FAILED;
    }

    if (!Utils::validStatus("aclrtSetDevice", aclrtSetDevice(deviceId_))) {
        ERROR_LOG("DeviceId is %d", deviceId_);
        (void)aclFinalize();
        return FAILED;
    }
    const char* soc_name{ aclrtGetSocName() };
    if (soc_name == nullptr)
    {
        ERROR_LOG("Can not get the soc name");
    }
    INFO_LOG("soc name: %s", soc_name);

    // create context (set current)
    if (!Utils::validStatus("aclrtCreateContext", aclrtCreateContext(&context_, deviceId_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // create stream
    if (!Utils::validStatus("aclrtCreateStream", aclrtCreateStream(&stream_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // get run mode
    // runMode is ACL_HOST which represents app is running in host
    // runMode is ACL_DEVICE which represents app is running in device
    aclrtRunMode runMode;
    if (!Utils::validStatus("aclrtGetRunMode", aclrtGetRunMode(&runMode))) {
        return FAILED;
    }
    g_isDevice = (runMode == ACL_DEVICE);
    return SUCCESS;
}

Result SampleProcess::Process()
{
    ModelProcess modelProcess;

    std::string pcds = "/home/mdc/mohsen/resnet50_11/data/point_cloud_data.pcd.bin" ;

    // model preprocessing
 
    preprocess_func(pcds);

    // model init
    
    const char* omModelPath = "/home/mdc/mohsen/resnet50_11/bin/model/pointpillars_opset_11_nonezero_custom_int32_simplified.om";
    Result ret = modelProcess.LoadModel(omModelPath);
    if (ret != SUCCESS) {
        ERROR_LOG("execute LoadModel failed");
        return FAILED;
    }

    ret = modelProcess.CreateModelDesc();
    if (ret != SUCCESS) {
        ERROR_LOG("execute CreateModelDesc failed");
        return FAILED;
    }


    size_t devBufferSize;
    void *picDevBuffer = nullptr;
    ret = modelProcess.GetInputSizeByIndex(0, devBufferSize);
    if (ret != SUCCESS) {
        ERROR_LOG("execute GetInputSizeByIndex failed");
        return FAILED;
    }

    aclError aclRet = aclrtMalloc(&picDevBuffer, devBufferSize, ACL_MEM_MALLOC_NORMAL_ONLY);
    if (!Utils::validStatus("aclrtMalloc", aclRet)) {
        ERROR_LOG("input for model size is %zu", devBufferSize);
        return FAILED;
    }


    size_t devBufferSize2;
    void *picDevBuffer2 = nullptr;
    ret = modelProcess.GetInputSizeByIndex(1, devBufferSize2);
    if (ret != SUCCESS) {
        ERROR_LOG("execute GetInputSizeByIndex failed");
        return FAILED;
    }

    aclError aclRet2 = aclrtMalloc(&picDevBuffer2, devBufferSize2, ACL_MEM_MALLOC_NORMAL_ONLY);
    if (!Utils::validStatus("aclrtMalloc", aclRet2)) {
        ERROR_LOG("input for model size is %zu", devBufferSize2);
        return FAILED;
    }


    size_t devBufferSize3;
    void *picDevBuffer3 = nullptr;
    ret = modelProcess.GetInputSizeByIndex(2, devBufferSize3);
    if (ret != SUCCESS) {
        ERROR_LOG("execute GetInputSizeByIndex failed");
        return FAILED;
    }

    aclError aclRet3 = aclrtMalloc(&picDevBuffer3, devBufferSize3, ACL_MEM_MALLOC_NORMAL_ONLY);
    if (!Utils::validStatus("aclrtMalloc", aclRet3)) {
        ERROR_LOG("input for model size is %zu", devBufferSize3);
        return FAILED;
    }

    vector <DataInfo> inputData = {{picDevBuffer, devBufferSize}, {picDevBuffer2, devBufferSize2}, {picDevBuffer3, devBufferSize3}};



    ret = modelProcess.CreateInput(inputData);
    if (ret != SUCCESS) {
        ERROR_LOG("execute CreateInput failed");
        aclrtFree(picDevBuffer);
        return FAILED;
    }




    ret = modelProcess.CreateOutput();
    if (ret != SUCCESS) {
        aclrtFree(picDevBuffer);
        ERROR_LOG("execute CreateOutput failed");
        return FAILED;
    }



    WARN_LOG("start to prepare inputs of model from preprocess");
    ret = Utils::MemcpyVariableToDeviceBuffer(num_points_data32, picDevBuffer, devBufferSize);
    Result ret1 = Utils::MemcpyVariableToDeviceBuffer(voxels_data, picDevBuffer2, devBufferSize2);
    Result ret2 = Utils::MemcpyVariableToDeviceBuffer(padded_coords_data32, picDevBuffer3, devBufferSize3);
    

        if (ret != SUCCESS) {
            aclrtFree(picDevBuffer);
            ERROR_LOG("memcpy device buffer failed, index is %d", 0);
            return FAILED;
        }
        if (ret1 != SUCCESS) {
            aclrtFree(picDevBuffer2);
            ERROR_LOG("memcpy device buffer failed, index is %d", 0);
            return FAILED;
        }
        if (ret2 != SUCCESS) {
            aclrtFree(picDevBuffer3);
            ERROR_LOG("memcpy device buffer failed, index is %d", 0);
            return FAILED;
        }
        
        ret = modelProcess.Execute();
        
        if (ret != SUCCESS) {
            ERROR_LOG("execute inference failed");
            aclrtFree(picDevBuffer);
            aclrtFree(picDevBuffer2);
            aclrtFree(picDevBuffer3);
            return FAILED;
        }

    // Print the top 5 confidence values with indexes.
    // Use function [DumpModelOutputResult] if you want to dump results to file in the current directory.
    auto model_output = modelProcess.OutputModelResult();
    
    auto final_results = postprocess_func(model_output);

    Matrix_R<float> enlarged_bboxes = enlargeBoundingBoxes<float>(final_results.bboxes, 0.1f);
    Matrix_R<float> minmax_bboxes = convertToMinMax<float>(enlarged_bboxes);
    //std::vector<BoundingBox> PointPillarResult = convertToVector(minmax_bboxes,final_results.labels,final_results.scores);
    std::vector<LidarPerceptionOutput> PointPillarResult = convertToVector(minmax_bboxes, final_results.labels,final_results.scores);
    /////////////////////Rule_based//////////////

    //Input of the rule_based
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    //pcl::io::loadPCDFile("../dataset/n008-2018-08-01-15-16-36-0400__LIDAR_TOP__1533151603547590.pcd.pcd", *cloud);
    pcl::io::loadPCDFile("/home/mdc/mohsen/resnet50_11/data/point_cloud_data.pcd.pcd", *cloud);
    
    LidarPerceptionInput input;
    // Need subscriber to feed the input sructure in every frame



    // Convert LidarPerceptionInput to PCL point cloud
    //pcl::PointCloud<pcl::PointXYZ> pclCloud = Lidar_processing::convertToPCL(input);
    

    // Use functions from lidar_processing namespace
    //Do Voxel Grid filtering
    auto voxel_filtered_cloud = Lidar_processing::VoxelGridFilter(cloud, 0.3);
    pcl::PointXYZ min_point, max_point;

    //Get Min and Max of PCD dimension
    Lidar_processing::getMinMaxParallel(*voxel_filtered_cloud, min_point, max_point);

    double X = max_point.x - min_point.x;
    double Y = max_point.y - min_point.y;
    double Z = max_point.z - min_point.z;

    // Define filter range using the computed bounding box dimensions
    Eigen::Vector4f min_range(-X / 2.0, -Y / 2.0, -Z / 2.0, 1.0);
    Eigen::Vector4f max_range(X / 2.0, Y / 2.0, Z / 2.0, 1.0);


    
    // Apply the filters
    auto filtered_cloud = Lidar_processing::filterRange(voxel_filtered_cloud, min_range, max_range);
    filtered_cloud = Lidar_processing::filterRoadBoundary(filtered_cloud, -10, 10);
    filtered_cloud = Lidar_processing::filterGround(filtered_cloud, 0.2);

    // Do euclideanClustering for detect objects in pcd
    //std::vector<BoundingBox> RuleBasedResult = Lidar_processing::euclideanClustering(filtered_cloud);
    std::vector<LidarPerceptionOutput> RuleBasedResult = Lidar_processing::euclideanClustering(filtered_cloud);
    

            /////////////////////Fusion//////////////
    // Calculate matching cost matrix with normalization
std::vector<std::vector<double>> costMatrix(RuleBasedResult.size(), std::vector<double>(PointPillarResult.size(), 0.0));
double maxCost = 0.0;

// Calculate matching cost using the specified criteria and find the maximum cost
for (size_t i = 0; i < RuleBasedResult.size(); ++i) {
    for (size_t j = 0; j < PointPillarResult.size(); ++j) {
        double distanceCost = calculateDistance(RuleBasedResult[i], PointPillarResult[j]);
        double iouCost = calculateIoU(RuleBasedResult[i], PointPillarResult[j]);
        double cost = 0.6 * distanceCost + 0.4 * iouCost;

        costMatrix[i][j] = cost;

        // Update the maximum cost
        maxCost = std::max(maxCost, cost);
    }
}

// Normalize the cost matrix to be between 0 and 1
if (maxCost > 0.0) {
    for (size_t i = 0; i < RuleBasedResult.size(); ++i) {
        for (size_t j = 0; j < PointPillarResult.size(); ++j) {
            costMatrix[i][j] /= maxCost;
        }
    }
}
    


    // Find the rule-based objects with high matching values and remove them
    const double threshold = 0.05;  // threshold
    std::vector<size_t> toRemove;
    for (size_t i = 0; i < RuleBasedResult.size(); ++i) {
        auto maxCostIter = std::min_element(costMatrix[i].begin(), costMatrix[i].end());
        if (*maxCostIter < threshold) {
            // Rule-based object has a high matching value, mark for removal
            toRemove.push_back(i);
        }
    }

    // Remove rule-based objects with high matching values
    for (auto it = toRemove.rbegin(); it != toRemove.rend(); ++it) {
        RuleBasedResult.erase(RuleBasedResult.begin() + *it);
    }

    // Append remaining rule-based objects to the list of model-based objects
    PointPillarResult.insert(PointPillarResult.end(), RuleBasedResult.begin(), RuleBasedResult.end());

    // Print or further process the updated model-based objects list
    for (const auto& result : PointPillarResult) {
        std::cout << "Model-Based BBox 2D: "
                  << result.bbox.xmin << " " << result.bbox.ymin << " "
                  << result.bbox.xmax << " " << result.bbox.ymax <<" "<<result.confidence<<" "<< std::endl;
    }
    





    // release model input output
    modelProcess.DestroyInput();
    modelProcess.DestroyOutput();

    aclrtFree(picDevBuffer);
    aclrtFree(picDevBuffer2);
    aclrtFree(picDevBuffer3);
    return SUCCESS;
}

void SampleProcess::DestroyResource()
{
    aclError ret;
    if (stream_ != nullptr) {
        ret = aclrtDestroyStream(stream_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy stream failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        stream_ = nullptr;
    }
    INFO_LOG("end to destroy stream");

    if (context_ != nullptr) {
        ret = aclrtDestroyContext(context_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy context failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        context_ = nullptr;
    }
    INFO_LOG("end to destroy context");

    ret = aclrtResetDevice(deviceId_);
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("reset device %d failed, errorCode = %d", deviceId_, static_cast<int32_t>(ret));
    }
    INFO_LOG("end to reset device %d", deviceId_);

    ret = aclFinalize();
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("finalize acl failed, errorCode = %d", static_cast<int32_t>(ret));
    }
    INFO_LOG("end to finalize acl");
}
